import re
import json
import logging
import requests
from decimal import Decimal
from PySide6.QtCore import Qt, QDateTime, QTimeZone
from PySide6.QtWidgets import QDialog
from jal.data_import.receipt_api.receipt_api import ReceiptAPI
from jal.db.settings import JalSettings
from jal.ui.ui_login_pingo_doce_dlg import Ui_LoginPingoDoceDialog


#-----------------------------------------------------------------------------------------------------------------------
class ReceiptPtPingoDoce(ReceiptAPI):
    receipt_pattern = r"A:.*\*B:.*\*C:PT\*D:FS\*E:N\*F:(?P<date>\d{8})\*G:FS (?P<shop_id>\d{4})(?P<register_id>\d{3}).*\/.*\*H:.{1,70}\*I1:PT\*.*\*O:(?P<amount>\d{1,}\.\d\d)\*.*"
    # aux data is in form SSSSSSYYYYMMDDhhmmNNNNCCCCMMMM
    # SSSSSS - receipt sequence number
    # YYYY, MM, DD, hh, mm - year, month, day and hour/minute of the receipt
    # NNNN - unknown 4 digits
    # CCCC cash register ID in the shop ID MMMM
    def __init__(self, qr_text='', aux_data='', params=None):
        super().__init__()
        self.aux_data = aux_data
        self.access_token = ''
        self.user_profile = {}
        self.receipts = []
        self.slip_json = {}
        if params is None:
            parts = re.match(self.receipt_pattern, qr_text)
            if parts is None:
                raise ValueError(self.tr("Pingo Doce QR available but pattern isn't recognized: " + qr_text))
            parts = parts.groupdict()
            self.date_time = QDateTime.fromString(parts['date'], 'yyyyMMdd')
            self.shop_id = int(parts['shop_id'].lstrip('0'))
            self.total_amount = float(parts['amount'])
        else:
            self.date_time = params['Date/Time']
            self.shop_id = params['Shop #']
            self.total_amount = float(params['Total'])
        self.web_session = requests.Session()
        self.web_session.headers['User-Agent'] = "okhttp/4.10.0"
        self.web_session.headers['Content-Type'] = 'application/json; charset=UTF-8'

    @staticmethod
    def parameters_list() -> dict:
        parameters = {
            "Date/Time": QDateTime.currentDateTime(QTimeZone.UTC),
            "Shop #": 0,
            "Total": Decimal('0')
        }
        return parameters

    def activate_session(self) -> bool:
        self.access_token = JalSettings().getValue('PtPingoDoceAccessToken', default='')
        self.user_profile = json.loads(JalSettings().getValue('PtPingoDoceUserProfile', default='{}'))
        if not self.access_token:
            self.__do_login()
        if not self.access_token or not self.user_profile:
            logging.warning(self.tr("No Pingo Doce access token available"))
            return False
        self.web_session.headers["Authorization"] = f"Bearer {self.access_token}"
        self.web_session.headers["pdapp-storeId"] = "-1"
        self.web_session.headers["pdapp-cardNumber"] = f"PDOM|{self.user_profile['ompdCard']}|{self.user_profile['householdId']}"
        self.web_session.headers["pdapp-lcid"] = self.user_profile['loyaltyId']
        self.web_session.headers["pdapp-hid"] = self.user_profile['householdId']
        payload = '{"pmCard":"' + self.user_profile['ompdCard'] + '"}'
        response = self.web_session.post("https://app-proxy.pingodoce.pt/api/v2/user/cardassociations/savings", data=payload)
        if response.status_code == 401:
            logging.info(self.tr("Unauthorized with reason: ") + f"{response.text}")
            if self.__refresh_token():
                self.web_session.headers["Authorization"] = f"Bearer {self.access_token}"
                self.web_session.headers["pdapp-storeId"] = "-1"
                self.web_session.headers["pdapp-cardNumber"] = f"PDOM|{self.user_profile['ompdCard']}|{self.user_profile['householdId']}"
                self.web_session.headers["pdapp-lcid"] = self.user_profile['loyaltyId']
                self.web_session.headers["pdapp-hid"] = self.user_profile['householdId']
            else:
                return False
        elif response.status_code != 200:
            logging.error(self.tr("Pingo Doce API failed with: ") + f"{response.status_code}/{response.text}")
            return False
        response = self.web_session.get("https://app-proxy.pingodoce.pt/api/v2/user/transactionsHistory/chart?filter=FILTER_BY_30_DAYS")
        if response.status_code != 200:
            logging.error(self.tr("Pingo Doce API filter failed with: ") + f"{response.status_code}/{response.text}")
            return False
        page = 1
        while page>0:
            logging.info(f"Loading {page} page of Pingo Doce receipts")
            response = self.web_session.get(f"https://app-proxy.pingodoce.pt/api/v2/user/transactionsHistory?pageNumber={page}&pageSize=20")
            if response.status_code != 200:
                logging.error(self.tr("Pingo Doce API history failed: ") + f"{response.status_code}/{response.text}")
                return False
            receipts = json.loads(response.text)
            page = -1 if len(receipts) < 20 else page + 1  # Go to next page or stop loading
            for receipt in receipts:
                self.receipts.append({"id": receipt['transactionId'], "shop_id": receipt['storeId'],
                                      "date": receipt['transactionDate'], "amount": float(receipt['total'])})
        logging.info(f"Pingo Doce list of receipts loaded: {self.receipts}")
        return True

    def __refresh_token(self) -> bool:
        logging.info(self.tr("Refreshing Pingo Doce token..."))
        self.access_token = JalSettings().getValue('PtPingoDoceAccessToken', default='')
        refresh_token = JalSettings().getValue('PtPingoDoceRefreshToken', default='')
        self.web_session = requests.Session()   # Need to start a clean session here
        self.web_session.headers['User-Agent'] = "okhttp/4.10.0"
        self.web_session.headers["Authorization"] = f"Bearer {self.access_token}"
        payload = {"client_id": "pdappclient", "grant_type": "refresh_token", "refresh_token": refresh_token}
        response = self.web_session.post("https://app-proxy.pingodoce.pt/connect/token", data=payload)
        if response.status_code == 200:
            logging.info(self.tr("Pingo Doce token was refreshed: ") + f"{response.text}")
            json_content = json.loads(response.text)
            assert json_content['token_type'] == "Bearer"
            self.access_token = json_content['access_token']
            new_refresh_token = json_content['refresh_token']
            settings = JalSettings()
            settings.setValue('PtPingoDoceAccessToken', self.access_token)
            settings.setValue('PtPingoDoceRefreshToken', new_refresh_token)
        else:
            logging.error(self.tr("Can't refresh Pingo Doce token, response: ") + f"{response.status_code}/{response.text}")
            JalSettings().setValue('PtPingoDoceAccessToken', '')
            self.access_token = ''
            return False
        response = self.web_session.get("https://app-proxy.pingodoce.pt/api/v2/user/userprofiles")
        if response.status_code != 200:
            logging.error(self.tr("Can't get Pingo Doce profile, response: ") + f"{response.status_code}/{response.text}")
            return False
        logging.info(self.tr("Pingo Doce profile was loaded: ") + f"{response.text}")
        self.user_profile = json.loads(response.text)
        settings.setValue('PtPingoDoceUserProfile', json.dumps(self.user_profile))
        return True

    def __do_login(self):
        login_dialog = LoginPingoDoce()
        if login_dialog.exec() == QDialog.Accepted:
            self.access_token = JalSettings().getValue('PtPingoDoceAccessToken')

    def query_slip(self):
        tickets = [x for x in self.receipts if x['shop_id'] == self.shop_id and
                   x['amount'] == self.total_amount and x['date'] == self.date_time.toString('yyyy-MM-ddT00:00:00')]
        if len(tickets) == 1:
            response = self.web_session.get(f"https://app-proxy.pingodoce.pt/api/v2/user/transactionsHistory/details?id={tickets[0]['id']}")
            if response.status_code == 200:
                logging.info(self.tr("Receipt was loaded: " + response.text))
                self.slip_json = json.loads(response.text)
                self.slip_load_ok.emit()
            else:
                logging.error(self.tr("Receipt load failed: ") + f"{response.status_code}/{response.text} for {tickets[0]['id']}")
                self.slip_json = {}
                self.slip_load_failed.emit()
        else:
            if len(tickets) == 0:
                logging.warning(self.tr("Receipt was not found in available list"))
            else:
                logging.warning(self.tr("Several similar receipts was found: ") + {tickets})
            self.slip_json = {}
            self.slip_load_failed.emit()

    def shop_name(self) -> str:
        return "Pingo Doce"

    def datetime(self) -> QDateTime:
        receipt_datetime = self.date_time
        receipt_datetime.setTimeSpec(Qt.UTC)
        return receipt_datetime

    def slip_lines(self) -> list:
        lines = self.slip_json['products']
        for line in lines:
            line['quantity'] = float(line.pop('purchaseQuantity').replace(',', '.'))
            line['amount'] = -float(line.pop('purchasePrice').replace(',', '.'))
            line['unit_price'] = abs(line['amount'] / line['quantity'])
            if line['quantity'] != 1:
                line['name'] = f"{line['name']} ({line['quantity']:g} x {line['unit_price']:.2f})"
        return lines


#-----------------------------------------------------------------------------------------------------------------------
class LoginPingoDoce(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_LoginPingoDoceDialog()
        self.ui.setupUi(self)
        self.ui.LoginBtn.pressed.connect(self.do_login)
        self.phone_number = ''
        self.web_session = requests.Session()
        self.web_session.headers['User-Agent'] = 'okhttp/4.2.2'

    def do_login(self):
        self.phone_number = self.ui.PhoneNumberEdit.text().replace('-', '')[4:]  # Take phone number without country code
        self.web_session.headers['Content-Type'] = 'application/json; charset=UTF-8'
        payload = '{' + f'"phoneNumber":"{self.phone_number}"' + '}'
        response = self.web_session.post("https://app-proxy.pingodoce.pt/api/v2/identity/sms/verifyNumber", data=payload)
        if response.status_code != 200:
            logging.error(self.tr("Pingo Doce phone verification failed: ") + f"{response.status_code}/{response.text}")
            return
        data = json.loads(response.text)
        terms = {'version': '', 'type': '', 'privacyUrl': '', 'termsOfUseUrl': '', 'title': ''}
        if data['status'] == 220:
            if data['message'] != "Utilizador deve aceitar a última versão dos consentimentos":
                logging.warning(f"Unusual message received: {data['message']}")
            terms.update(data['consents'])
        elif data['status'] == 200:
            pass
        else:
            logging.error(self.tr("Pingo Doce unknown login status: ") + f"{data['status']} / {data['message']}")
            return
        login_data = '{"phoneNumber":"' + self.phone_number + '","password":"' + self.ui.PasswordEdit.text() \
                     + '","consents":' + json.dumps(terms) + '}'
        response = self.web_session.post("https://app-proxy.pingodoce.pt/api/v2/identity/onboarding/login",
                                         data=login_data)
        if response.status_code != 200:
            logging.error(self.tr("Pingo Doce login failed: ") + f"{response.status_code}/{response.text}")
            return
        logging.info(self.tr("Pingo Doce login successful: ") + f"{response.text}")
        json_content = json.loads(response.text)
        try:
            assert json_content['token']['token_type'] == "Bearer"
            access_token = json_content['token']['access_token']
            refresh_token = json_content['token']['refresh_token']
            user_profile = json_content['profile']
            settings = JalSettings()
            settings.setValue('PtPingoDoceAccessToken', access_token)
            settings.setValue('PtPingoDoceRefreshToken', refresh_token)
            settings.setValue('PtPingoDoceUserProfile', json.dumps(user_profile))
            self.accept()
        except KeyError as e:
            logging.error(self.tr("Pingo Doce login response failed with: ") + e)
            return



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'portfolio_report.ui'
##
## Created by: Qt User Interface Compiler version 6.6.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QComboBox, QDateEdit,
    QFrame, QHBoxLayout, QHeaderView, QLabel,
    QPushButton, QSizePolicy, QSpacerItem, QTreeView,
    QVBoxLayout, QWidget)

from jal.widgets.account_select import CurrencyComboBox

class Ui_PortfolioWidget(object):
    def setupUi(self, PortfolioWidget):
        if not PortfolioWidget.objectName():
            PortfolioWidget.setObjectName(u"PortfolioWidget")
        PortfolioWidget.resize(1066, 589)
        self.verticalLayout = QVBoxLayout(PortfolioWidget)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.PortfolioParamsFrame = QFrame(PortfolioWidget)
        self.PortfolioParamsFrame.setObjectName(u"PortfolioParamsFrame")
        self.PortfolioParamsFrame.setFrameShape(QFrame.Panel)
        self.PortfolioParamsFrame.setFrameShadow(QFrame.Sunken)
        self.horizontalLayout_8 = QHBoxLayout(self.PortfolioParamsFrame)
        self.horizontalLayout_8.setSpacing(6)
        self.horizontalLayout_8.setObjectName(u"horizontalLayout_8")
        self.horizontalLayout_8.setContentsMargins(2, 2, 2, 2)
        self.PortfolioDate = QDateEdit(self.PortfolioParamsFrame)
        self.PortfolioDate.setObjectName(u"PortfolioDate")
        self.PortfolioDate.setDateTime(QDateTime(QDate(2020, 11, 24), QTime(0, 0, 0)))
        self.PortfolioDate.setCalendarPopup(True)
        self.PortfolioDate.setTimeSpec(Qt.UTC)

        self.horizontalLayout_8.addWidget(self.PortfolioDate)

        self.GroupLbl = QLabel(self.PortfolioParamsFrame)
        self.GroupLbl.setObjectName(u"GroupLbl")

        self.horizontalLayout_8.addWidget(self.GroupLbl)

        self.GroupCombo = QComboBox(self.PortfolioParamsFrame)
        self.GroupCombo.setObjectName(u"GroupCombo")

        self.horizontalLayout_8.addWidget(self.GroupCombo)

        self.PortfolioCurrencyLbl = QLabel(self.PortfolioParamsFrame)
        self.PortfolioCurrencyLbl.setObjectName(u"PortfolioCurrencyLbl")

        self.horizontalLayout_8.addWidget(self.PortfolioCurrencyLbl)

        self.PortfolioCurrencyCombo = CurrencyComboBox(self.PortfolioParamsFrame)
        self.PortfolioCurrencyCombo.setObjectName(u"PortfolioCurrencyCombo")

        self.horizontalLayout_8.addWidget(self.PortfolioCurrencyCombo)

        self.ShowInactiveAccounts = QCheckBox(self.PortfolioParamsFrame)
        self.ShowInactiveAccounts.setObjectName(u"ShowInactiveAccounts")

        self.horizontalLayout_8.addWidget(self.ShowInactiveAccounts)

        self.horizontalSpacer = QSpacerItem(1411, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_8.addItem(self.horizontalSpacer)

        self.SaveButton = QPushButton(self.PortfolioParamsFrame)
        self.SaveButton.setObjectName(u"SaveButton")

        self.horizontalLayout_8.addWidget(self.SaveButton)


        self.verticalLayout.addWidget(self.PortfolioParamsFrame)

        self.PortfolioTreeView = QTreeView(PortfolioWidget)
        self.PortfolioTreeView.setObjectName(u"PortfolioTreeView")
        self.PortfolioTreeView.setFrameShape(QFrame.Panel)
        self.PortfolioTreeView.setAlternatingRowColors(True)
        self.PortfolioTreeView.setAnimated(True)
        self.PortfolioTreeView.setAllColumnsShowFocus(True)

        self.verticalLayout.addWidget(self.PortfolioTreeView)


        self.retranslateUi(PortfolioWidget)

        QMetaObject.connectSlotsByName(PortfolioWidget)
    # setupUi

    def retranslateUi(self, PortfolioWidget):
        PortfolioWidget.setWindowTitle(QCoreApplication.translate("PortfolioWidget", u"Asset Portfolio", None))
        self.PortfolioDate.setDisplayFormat(QCoreApplication.translate("PortfolioWidget", u"dd/MM/yyyy", None))
        self.GroupLbl.setText(QCoreApplication.translate("PortfolioWidget", u"Group by:", None))
        self.PortfolioCurrencyLbl.setText(QCoreApplication.translate("PortfolioWidget", u"Common currency:", None))
        self.ShowInactiveAccounts.setText(QCoreApplication.translate("PortfolioWidget", u"Show &Inactive accounts", None))
        self.SaveButton.setText(QCoreApplication.translate("PortfolioWidget", u"Save...", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'receipt_import_dlg.ui'
##
## Created by: Qt User Interface Compiler version 6.6.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QDateTimeEdit, QDialog,
    QFrame, QGridLayout, QGroupBox, QHBoxLayout,
    QHeaderView, QLabel, QLineEdit, QPushButton,
    QSizePolicy, QSpacerItem, QTableView, QVBoxLayout,
    QWidget)

from jal.widgets.reference_selector import (AccountSelector, PeerSelector)

class Ui_ImportShopReceiptDlg(object):
    def setupUi(self, ImportShopReceiptDlg):
        if not ImportShopReceiptDlg.objectName():
            ImportShopReceiptDlg.setObjectName(u"ImportShopReceiptDlg")
        ImportShopReceiptDlg.resize(850, 587)
        self.verticalLayout = QVBoxLayout(ImportShopReceiptDlg)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(2, 2, 2, 2)
        self.InputFrame = QFrame(ImportShopReceiptDlg)
        self.InputFrame.setObjectName(u"InputFrame")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.InputFrame.sizePolicy().hasHeightForWidth())
        self.InputFrame.setSizePolicy(sizePolicy)
        self.InputFrame.setFrameShape(QFrame.NoFrame)
        self.InputFrame.setFrameShadow(QFrame.Plain)
        self.horizontalLayout_2 = QHBoxLayout(self.InputFrame)
        self.horizontalLayout_2.setSpacing(2)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.QRGroup = QGroupBox(self.InputFrame)
        self.QRGroup.setObjectName(u"QRGroup")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.QRGroup.sizePolicy().hasHeightForWidth())
        self.QRGroup.setSizePolicy(sizePolicy1)
        self.QRGroup.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignVCenter)
        self.horizontalLayout = QHBoxLayout(self.QRGroup)
        self.horizontalLayout.setSpacing(6)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(2, 2, 2, 2)
        self.ScanReceiptQR = QPushButton(self.QRGroup)
        self.ScanReceiptQR.setObjectName(u"ScanReceiptQR")

        self.horizontalLayout.addWidget(self.ScanReceiptQR)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)


        self.horizontalLayout_2.addWidget(self.QRGroup)


        self.verticalLayout.addWidget(self.InputFrame)

        self.SlipDataGroup = QGroupBox(ImportShopReceiptDlg)
        self.SlipDataGroup.setObjectName(u"SlipDataGroup")
        self.SlipDataGroup.setEnabled(True)
        sizePolicy1.setHeightForWidth(self.SlipDataGroup.sizePolicy().hasHeightForWidth())
        self.SlipDataGroup.setSizePolicy(sizePolicy1)
        self.gridLayout_2 = QGridLayout(self.SlipDataGroup)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setContentsMargins(2, 2, 2, 2)
        self.ReceiptAPICombo = QComboBox(self.SlipDataGroup)
        self.ReceiptAPICombo.setObjectName(u"ReceiptAPICombo")

        self.gridLayout_2.addWidget(self.ReceiptAPICombo, 0, 1, 1, 1)

        self.ReceiptAPILabel = QLabel(self.SlipDataGroup)
        self.ReceiptAPILabel.setObjectName(u"ReceiptAPILabel")

        self.gridLayout_2.addWidget(self.ReceiptAPILabel, 0, 0, 1, 1)

        self.DownloadReceiptBtn = QPushButton(self.SlipDataGroup)
        self.DownloadReceiptBtn.setObjectName(u"DownloadReceiptBtn")

        self.gridLayout_2.addWidget(self.DownloadReceiptBtn, 2, 0, 1, 2)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.gridLayout_2.addItem(self.verticalSpacer, 1, 0, 1, 1)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.gridLayout_2.addItem(self.horizontalSpacer_2, 0, 3, 1, 1)

        self.ReceiptParametersList = QTableView(self.SlipDataGroup)
        self.ReceiptParametersList.setObjectName(u"ReceiptParametersList")
        self.ReceiptParametersList.horizontalHeader().setVisible(False)
        self.ReceiptParametersList.horizontalHeader().setStretchLastSection(True)
        self.ReceiptParametersList.verticalHeader().setMinimumSectionSize(20)
        self.ReceiptParametersList.verticalHeader().setDefaultSectionSize(20)

        self.gridLayout_2.addWidget(self.ReceiptParametersList, 0, 2, 3, 1)


        self.verticalLayout.addWidget(self.SlipDataGroup)

        self.ReceiptGroup = QGroupBox(ImportShopReceiptDlg)
        self.ReceiptGroup.setObjectName(u"ReceiptGroup")
        sizePolicy2 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.ReceiptGroup.sizePolicy().hasHeightForWidth())
        self.ReceiptGroup.setSizePolicy(sizePolicy2)
        self.gridLayout = QGridLayout(self.ReceiptGroup)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(2, 2, 2, 2)
        self.SlipDateTime = QDateTimeEdit(self.ReceiptGroup)
        self.SlipDateTime.setObjectName(u"SlipDateTime")
        self.SlipDateTime.setTimeSpec(Qt.UTC)

        self.gridLayout.addWidget(self.SlipDateTime, 2, 1, 1, 1)

        self.DateTimeLbl = QLabel(self.ReceiptGroup)
        self.DateTimeLbl.setObjectName(u"DateTimeLbl")

        self.gridLayout.addWidget(self.DateTimeLbl, 1, 1, 1, 1)

        self.CorrespondenceLbl = QLabel(self.ReceiptGroup)
        self.CorrespondenceLbl.setObjectName(u"CorrespondenceLbl")

        self.gridLayout.addWidget(self.CorrespondenceLbl, 3, 2, 1, 1)

        self.PeerEdit = PeerSelector(self.ReceiptGroup)
        self.PeerEdit.setObjectName(u"PeerEdit")

        self.gridLayout.addWidget(self.PeerEdit, 3, 3, 1, 1)

        self.PeerLbl = QLabel(self.ReceiptGroup)
        self.PeerLbl.setObjectName(u"PeerLbl")

        self.gridLayout.addWidget(self.PeerLbl, 3, 0, 1, 1)

        self.SlipShopName = QLineEdit(self.ReceiptGroup)
        self.SlipShopName.setObjectName(u"SlipShopName")
        self.SlipShopName.setEnabled(False)

        self.gridLayout.addWidget(self.SlipShopName, 3, 1, 1, 1)

        self.LinesLbl = QLabel(self.ReceiptGroup)
        self.LinesLbl.setObjectName(u"LinesLbl")
        self.LinesLbl.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)

        self.gridLayout.addWidget(self.LinesLbl, 4, 0, 1, 1)

        self.AccountLbl = QLabel(self.ReceiptGroup)
        self.AccountLbl.setObjectName(u"AccountLbl")

        self.gridLayout.addWidget(self.AccountLbl, 1, 3, 1, 1)

        self.AccountEdit = AccountSelector(self.ReceiptGroup)
        self.AccountEdit.setObjectName(u"AccountEdit")

        self.gridLayout.addWidget(self.AccountEdit, 2, 3, 1, 1)

        self.LinesTableView = QTableView(self.ReceiptGroup)
        self.LinesTableView.setObjectName(u"LinesTableView")
        self.LinesTableView.verticalHeader().setVisible(False)
        self.LinesTableView.verticalHeader().setMinimumSectionSize(20)
        self.LinesTableView.verticalHeader().setDefaultSectionSize(20)

        self.gridLayout.addWidget(self.LinesTableView, 4, 1, 1, 4)

        self.AssignCategoryBtn = QPushButton(self.ReceiptGroup)
        self.AssignCategoryBtn.setObjectName(u"AssignCategoryBtn")

        self.gridLayout.addWidget(self.AssignCategoryBtn, 2, 4, 1, 1)

        self.AssignTagBtn = QPushButton(self.ReceiptGroup)
        self.AssignTagBtn.setObjectName(u"AssignTagBtn")

        self.gridLayout.addWidget(self.AssignTagBtn, 3, 4, 1, 1)


        self.verticalLayout.addWidget(self.ReceiptGroup)

        self.DialogButtonsFrame = QFrame(ImportShopReceiptDlg)
        self.DialogButtonsFrame.setObjectName(u"DialogButtonsFrame")
        self.DialogButtonsFrame.setFrameShape(QFrame.NoFrame)
        self.DialogButtonsFrame.setFrameShadow(QFrame.Plain)
        self.horizontalLayout_4 = QHBoxLayout(self.DialogButtonsFrame)
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.horizontalLayout_4.setContentsMargins(2, 2, 2, 2)
        self.ClearBtn = QPushButton(self.DialogButtonsFrame)
        self.ClearBtn.setObjectName(u"ClearBtn")
        self.ClearBtn.setEnabled(True)

        self.horizontalLayout_4.addWidget(self.ClearBtn)

        self.AddOperationBtn = QPushButton(self.DialogButtonsFrame)
        self.AddOperationBtn.setObjectName(u"AddOperationBtn")
        self.AddOperationBtn.setEnabled(True)

        self.horizontalLayout_4.addWidget(self.AddOperationBtn)

        self.CloseBtn = QPushButton(self.DialogButtonsFrame)
        self.CloseBtn.setObjectName(u"CloseBtn")

        self.horizontalLayout_4.addWidget(self.CloseBtn)


        self.verticalLayout.addWidget(self.DialogButtonsFrame)


        self.retranslateUi(ImportShopReceiptDlg)
        self.CloseBtn.clicked.connect(ImportShopReceiptDlg.close)

        QMetaObject.connectSlotsByName(ImportShopReceiptDlg)
    # setupUi

    def retranslateUi(self, ImportShopReceiptDlg):
        ImportShopReceiptDlg.setWindowTitle(QCoreApplication.translate("ImportShopReceiptDlg", u"Import shop receipt", None))
        self.QRGroup.setTitle(QCoreApplication.translate("ImportShopReceiptDlg", u"Get receipt by scanning QR-code", None))
        self.ScanReceiptQR.setText(QCoreApplication.translate("ImportShopReceiptDlg", u"Scan receipt QR", None))
        self.SlipDataGroup.setTitle(QCoreApplication.translate("ImportShopReceiptDlg", u"Get receipt by manual data entry", None))
        self.ReceiptAPILabel.setText(QCoreApplication.translate("ImportShopReceiptDlg", u"Receipt type:", None))
        self.DownloadReceiptBtn.setText(QCoreApplication.translate("ImportShopReceiptDlg", u"Download receipt", None))
        self.ReceiptGroup.setTitle(QCoreApplication.translate("ImportShopReceiptDlg", u"Operation data", None))
        self.SlipDateTime.setDisplayFormat(QCoreApplication.translate("ImportShopReceiptDlg", u"dd/MM/yyyy hh:mm:ss", None))
        self.DateTimeLbl.setText(QCoreApplication.translate("ImportShopReceiptDlg", u"Date / Time:", None))
        self.CorrespondenceLbl.setText(QCoreApplication.translate("ImportShopReceiptDlg", u" \u279c ", None))
        self.PeerLbl.setText(QCoreApplication.translate("ImportShopReceiptDlg", u"Peer:", None))
        self.LinesLbl.setText(QCoreApplication.translate("ImportShopReceiptDlg", u"Lines:", None))
        self.AccountLbl.setText(QCoreApplication.translate("ImportShopReceiptDlg", u"Account:", None))
        self.AssignCategoryBtn.setText(QCoreApplication.translate("ImportShopReceiptDlg", u"Auto-assign categories", None))
        self.AssignTagBtn.setText(QCoreApplication.translate("ImportShopReceiptDlg", u"Set Tag for all lines", None))
        self.ClearBtn.setText(QCoreApplication.translate("ImportShopReceiptDlg", u"Clear", None))
        self.AddOperationBtn.setText(QCoreApplication.translate("ImportShopReceiptDlg", u"Add", None))
        self.CloseBtn.setText(QCoreApplication.translate("ImportShopReceiptDlg", u"Close", None))
    # retranslateUi




import re
import json
import logging
import requests
import requests_oauthlib
from urllib import parse
from oauthlib.oauth2 import MobileApplicationClient
from PySide6.QtCore import Qt, Slot, Signal, QMetaObject, QDateTime, QTimeZone, QUrl
from PySide6.QtWidgets import QDialog, QInputDialog, QMessageBox
from PySide6.QtWebEngineCore import QWebEngineProfile, QWebEnginePage, QWebEngineUrlScheme, \
    QWebEngineUrlSchemeHandler, QWebEngineUrlRequestJob
from jal.data_import.receipt_api.receipt_api import ReceiptAPI
from jal.db.settings import JalSettings
from jal.ui.ui_login_lidl_plus_dlg import Ui_LoginLidlPlusDialog


#-----------------------------------------------------------------------------------------------------------------------
class ReceiptEuLidlPlus(ReceiptAPI):
    receipt_pattern = r"A:.*\*B:.*\*C:PT\*D:FS\*E:N\*F:(?P<date>\d{8})\*G:FS (?P<shop_id>\d{4})\d(?P<register_id>\d{2})..\/.*\*H:.{1,70}\*I1:PT\*.*"
    # aux data is in form 888MMMMSSSSSSCCDDMMYYXFFFFFF
    # SSSSSS - receipt sequence number
    # YY, MM, DD - year, month, day of the receipt
    # CC cash register ID in the shop ID MMMM
    # X - unknown 1 digit
    # FFFFFF - FS part of fiscal data
    def __init__(self, qr_text='', aux_data='', params=None):
        super().__init__()
        self.access_token = ''
        self.slip_json = {}
        if params is None:
            parts = re.match(self.receipt_pattern, qr_text)
            if parts is None:
                raise ValueError(self.tr("Lidl QR available but pattern isn't recognized: " + qr_text))
            parts = parts.groupdict()
            self.date_time = QDateTime.fromString(parts['date'], 'yyyyMMdd')
            self.shop_id = int(parts['shop_id'])
            self.register_id = int(parts['register_id'])
            if len(aux_data) == 28:  # Get receipt sequence number from aux data or from the user
                self.seq_id = aux_data[7:13]
            else:
                self.seq_id, result = QInputDialog.getText(None, self.tr("Input Lidl receipt additional data"),
                                                           self.tr("Sequence #:"))
                if not result:
                    raise ValueError(self.tr("Can't get Lidl receipt without sequence number"))
            self.seq_id = int(self.seq_id.lstrip('0'))   # Get rid of any leading zeros and convert to int
        else:
            self.date_time = params['Date']
            self.shop_id = params['Shop #']
            self.register_id = params['Register #']
            self.seq_id = params['Sequence #']
        self.web_session = requests.Session()
        self.web_session.headers['User-Agent'] = "okhttp/4.10.0"

    @staticmethod
    def parameters_list() -> dict:
        parameters = {
            "Date": QDateTime.currentDateTime(QTimeZone.UTC).date(),
            "Shop #": 0,
            "Register #": 0,
            "Sequence #": 0
        }
        return parameters

    def activate_session(self) -> bool:
        self.access_token = JalSettings().getValue('EuLidlAccessToken', default='')
        if not self.access_token:
            self.__do_login()
        if not self.access_token:
            logging.warning(self.tr("No Lidl Plus access token available"))
            return False
        self.web_session.headers["Authorization"] = f"Bearer {self.access_token}"
        response = self.web_session.get("https://tickets.lidlplus.com/api/v2/PT/tickets?pageNumber=1&onlyFavorite=false&itemId=")  # Just to check authorization status
        if response.status_code == 200:
            return True
        if response.status_code == 401:
            logging.info(self.tr("Unauthorized with reason: ") + f"{response.text}")
            return self.__refresh_token()
        else:
            logging.error(self.tr("Lidl Plus API failed with: ") + f"{response.status_code}/{response.text}")
            return False

    def __refresh_token(self) -> bool:
        logging.info(self.tr("Refreshing Lidl Plus token..."))
        client_secret = JalSettings().getValue('EuLidlClientSecret')
        refresh_token = JalSettings().getValue('EuLidlRefreshToken')
        self.web_session.headers["Authorization"] = f"Basic {client_secret}"
        self.web_session.headers['Accept'] = "application/json"
        payload = {"grant_type": "refresh_token", "refresh_token": refresh_token}
        response = self.web_session.post("https://accounts.lidl.com/connect/token", data=payload)
        if response.status_code == 200:
            logging.info(self.tr("Lidl Plus token was refreshed: ") + f"{response.text}")
            json_content = json.loads(response.text)
            assert json_content['token_type'] == "Bearer"
            self.access_token = json_content['access_token']
            new_refresh_token = json_content['refresh_token']
            settings = JalSettings()
            settings.setValue('EuLidlAccessToken', self.access_token)
            settings.setValue('EuLidlRefreshToken', new_refresh_token)
            self.web_session.headers["Authorization"] = f"Bearer {self.access_token}"
            return True
        else:
            logging.error(self.tr("Can't refresh Lidl Plus token, response: ") + f"{response.status_code}/{response.text}")
            JalSettings().setValue('EuLidlAccessToken', '')
            self.access_token = ''
            return False

    def __do_login(self):
        login_dialog = LoginLidlPlus()
        if login_dialog.exec() == QDialog.Accepted:
            self.access_token = JalSettings().getValue('EuLidlAccessToken')

    def query_slip(self):
        ticket_id = f"2500{self.shop_id:04d}{self.register_id:02d}{self.date_time.toString('yyyyMMdd')}{self.seq_id:05d}"
        self.web_session.headers["Accept-Language"] = "PT"
        response = self.web_session.get(f"https://tickets.lidlplus.com/api/v2/PT/tickets/{ticket_id}")
        if response.status_code == 200:
            logging.info(self.tr("Receipt was loaded: " + response.text))
            self.slip_json = json.loads(response.text)
            self.slip_load_ok.emit()
        else:
            logging.error(self.tr("Receipt load failed: ") + f"{response.status_code}/{response.text} for {ticket_id}")
            self.slip_json = {}
            self.slip_load_failed.emit()

    def shop_name(self) -> str:
        return "Lidl"

    def datetime(self) -> QDateTime:
        if 'date' in self.slip_json:
            receipt_datetime = QDateTime.fromString(self.slip_json['date'], 'yyyy-MM-ddTHH:mm:ss')
        else:
            receipt_datetime = QDateTime()
        receipt_datetime.setTimeSpec(Qt.UTC)
        return receipt_datetime

    def slip_lines(self) -> list:
        lines = self.slip_json['itemsLine']
        for line in lines:
            line['quantity'] = float(line['quantity'].replace(',', '.'))
            line['unit_price'] = float(line.pop('currentUnitPrice').replace(',', '.'))
            line['amount'] = -float(line.pop('originalAmount').replace(',', '.'))
            if line['quantity'] != 1:
                line['name'] = f"{line['name']} ({line['quantity']:g} x {line['unit_price']:.2f})"
        return lines


#-----------------------------------------------------------------------------------------------------------------------
class LoginLidlPlus(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.running = False
        self.ui = Ui_LoginLidlPlusDialog()
        self.ui.setupUi(self)

        self.verifier = ''
        self.lidl_client_id = "LidlPlusNativeClient"
        # Create scheme, handler and assign it to web-profile that is used for browser widget
        self.app_lidl_uri_scheme = QWebEngineUrlScheme(appDataHanlder.scheme)
        self.app_lidl_uri_scheme.setSyntax(QWebEngineUrlScheme.Syntax.Path)
        self.app_lidl_uri_scheme.setFlags(
            QWebEngineUrlScheme.Flag.ContentSecurityPolicyIgnored | QWebEngineUrlScheme.Flag.LocalAccessAllowed)
        QWebEngineUrlScheme.registerScheme(self.app_lidl_uri_scheme)
        self.app_lidl_uri_handler = appDataHanlder(self)
        self.app_lidl_uri_handler.data_received.connect(self.processResponse)
        self.web_profile = QWebEngineProfile(self)
        self.web_profile.installUrlSchemeHandler(appDataHanlder.scheme, self.app_lidl_uri_handler)
        self.ui.LidlPlusWebView.setPage(QWebEnginePage(self.web_profile, self))

    @Slot()
    def showEvent(self, event):
        super().showEvent(event)
        if self.running:
            return
        self.running = True
        # Call slot via queued connection, so it's called from the UI thread after the window has been shown
        QMetaObject().invokeMethod(self, "loadLoginPage", Qt.ConnectionType.QueuedConnection)

    @Slot()
    def loadLoginPage(self):
        # Get Login URL and open it in browser
        client = MobileApplicationClient(client_id=self.lidl_client_id)
        self.verifier = client.create_code_verifier(86)
        challenge = client.create_code_challenge(self.verifier, code_challenge_method='S256')
        oauth = requests_oauthlib.OAuth2Session(client_id=self.lidl_client_id,
                                                scope='openid profile offline_access lpprofile lpapis')
        oauth_url, oauth_state = oauth.authorization_url(url='https://accounts.lidl.com/')
        url = f"https://accounts.lidl.com/Account/Login?ReturnUrl=%2Fconnect%2Fauthorize%2Fcallback%3Fredirect_uri%3Dcom.lidlplus.app%253A%252F%252Fcallback"
        url += f"%26client_id%3D{self.lidl_client_id}"
        url += f"%26response_type%3Dcode%26state%3D{oauth_state}"  # "%26nonce%3Dt5Z_q8gekXoFRNmOVq2Ufg" - this part was omitted
        url += f"%26scope%3Dopenid%2520profile%2520offline_access%2520lpprofile%2520lpapis"
        url += f"%26code_challenge%3D{challenge}%26code_challenge_method%3DS256%26language%3DPT-PT%26track%3Dfalse%26force%3Dfalse%26Country%3DPT"
        self.ui.LidlPlusWebView.load(QUrl(url))   # Country and language are hard-coded to Portugal

    @Slot()
    def processResponse(self, params: dict):
        if 'code' not in params:
            QMessageBox().warning(None, self.tr("Login error"), self.tr("No auth code in callback URI"), QMessageBox.Ok)
            return
        client_secret = JalSettings().getValue('EuLidlClientSecret')
        # Get Auth token and display it
        s = requests.Session()
        s.headers["Authorization"] = f"Basic {client_secret}"
        s.headers['Accept'] = "application/json"
        data = {
            "grant_type": "authorization_code",
            "code": params['code'],
            "redirect_uri": "com.lidlplus.app://callback",
            "code_verifier": self.verifier
        }
        response = s.post("https://accounts.lidl.com/connect/token", data=data)
        if response.status_code != 200:
            logging.error(self.tr("Lidl Plus login failed: ") + f"{response.status_code}/{response.text}")
            return
        logging.info(self.tr("Lidl Plus login successful: ") + f"{response.text}")
        json_content = json.loads(response.text)
        assert json_content['token_type'] == "Bearer"
        new_access_token = json_content['access_token']
        new_refresh_token = json_content['refresh_token']
        settings = JalSettings()
        settings.setValue('EuLidlAccessToken', new_access_token)
        settings.setValue('EuLidlRefreshToken', new_refresh_token)
        self.accept()


# ----------------------------------------------------------------------------------------------------------------------
# Final URI is returned in form of "com.lidlplus.app://callback?code=..&scope=..&state=..&session_state=..", so
# it requires a special handler for "protocol" "com.lidlplus.app"
# Signal "data_received" is emitted when URI is processed, its parameter contains dictionary with URI parameters
# Page is redirected to "about:blank" in order to clean the browser window
class appDataHanlder(QWebEngineUrlSchemeHandler):
    scheme = b"com.lidlplus.app"
    data_received = Signal(dict)

    def __init__(self, parent = None):
        super().__init__(parent)

    def requestStarted(self, job: QWebEngineUrlRequestJob) -> None:
        url = job.requestUrl().url()
        params = dict(parse.parse_qsl(parse.urlsplit(url).query))
        job.fail(QWebEngineUrlRequestJob.Error.NoError)   # Stop processing but without an error
        self.data_received.emit(params)  # Dict contains 'code', 'scope', 'state' and 'session_state' elements



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'income_spending_operation.ui'
##
## Created by: Qt User Interface Compiler version 6.6.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QDateTimeEdit, QGridLayout, QHeaderView,
    QLabel, QLineEdit, QPushButton, QSizePolicy,
    QSpacerItem, QWidget)

from jal.widgets.account_select import (AccountCurrencyLabel, OptionalCurrencyComboBox)
from jal.widgets.custom.tableview_with_footer import TableViewWithFooter
from jal.widgets.reference_selector import (AccountSelector, PeerSelector)

class Ui_IncomeSpendingOperation(object):
    def setupUi(self, IncomeSpendingOperation):
        if not IncomeSpendingOperation.objectName():
            IncomeSpendingOperation.setObjectName(u"IncomeSpendingOperation")
        IncomeSpendingOperation.resize(968, 231)
        self.layout = QGridLayout(IncomeSpendingOperation)
        self.layout.setObjectName(u"layout")
        self.layout.setContentsMargins(2, 2, 2, 2)
        self.note = QLineEdit(IncomeSpendingOperation)
        self.note.setObjectName(u"note")

        self.layout.addWidget(self.note, 2, 7, 1, 1)

        self.account_label = QLabel(IncomeSpendingOperation)
        self.account_label.setObjectName(u"account_label")

        self.layout.addWidget(self.account_label, 1, 4, 1, 1)

        self.date_label = QLabel(IncomeSpendingOperation)
        self.date_label.setObjectName(u"date_label")

        self.layout.addWidget(self.date_label, 1, 0, 1, 1)

        self.vertical_spacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.layout.addItem(self.vertical_spacer, 4, 0, 1, 1)

        self.add_button = QPushButton(IncomeSpendingOperation)
        self.add_button.setObjectName(u"add_button")

        self.layout.addWidget(self.add_button, 2, 1, 1, 1)

        self.del_button = QPushButton(IncomeSpendingOperation)
        self.del_button.setObjectName(u"del_button")

        self.layout.addWidget(self.del_button, 2, 3, 1, 1)

        self.copy_button = QPushButton(IncomeSpendingOperation)
        self.copy_button.setObjectName(u"copy_button")

        self.layout.addWidget(self.copy_button, 2, 2, 1, 1)

        self.account_widget = AccountSelector(IncomeSpendingOperation)
        self.account_widget.setObjectName(u"account_widget")

        self.layout.addWidget(self.account_widget, 1, 5, 1, 1)

        self.note_label = QLabel(IncomeSpendingOperation)
        self.note_label.setObjectName(u"note_label")

        self.layout.addWidget(self.note_label, 2, 6, 1, 1)

        self.commit_button = QPushButton(IncomeSpendingOperation)
        self.commit_button.setObjectName(u"commit_button")
        self.commit_button.setEnabled(False)

        self.layout.addWidget(self.commit_button, 0, 9, 1, 1)

        self.peer_label = QLabel(IncomeSpendingOperation)
        self.peer_label.setObjectName(u"peer_label")

        self.layout.addWidget(self.peer_label, 2, 4, 1, 1)

        self.timestamp_editor = QDateTimeEdit(IncomeSpendingOperation)
        self.timestamp_editor.setObjectName(u"timestamp_editor")
        self.timestamp_editor.setCalendarPopup(True)
        self.timestamp_editor.setTimeSpec(Qt.UTC)

        self.layout.addWidget(self.timestamp_editor, 1, 1, 1, 3)

        self.main_label = QLabel(IncomeSpendingOperation)
        self.main_label.setObjectName(u"main_label")
        font = QFont()
        font.setBold(True)
        self.main_label.setFont(font)

        self.layout.addWidget(self.main_label, 0, 0, 1, 1)

        self.details_label = QLabel(IncomeSpendingOperation)
        self.details_label.setObjectName(u"details_label")

        self.layout.addWidget(self.details_label, 2, 0, 1, 1)

        self.revert_button = QPushButton(IncomeSpendingOperation)
        self.revert_button.setObjectName(u"revert_button")
        self.revert_button.setEnabled(False)
        self.revert_button.setAcceptDrops(False)

        self.layout.addWidget(self.revert_button, 0, 10, 1, 1)

        self.peer_widget = PeerSelector(IncomeSpendingOperation)
        self.peer_widget.setObjectName(u"peer_widget")

        self.layout.addWidget(self.peer_widget, 2, 5, 1, 1)

        self.details_table = TableViewWithFooter(IncomeSpendingOperation)
        self.details_table.setObjectName(u"details_table")
        self.details_table.setAlternatingRowColors(True)
        self.details_table.verticalHeader().setVisible(False)
        self.details_table.verticalHeader().setMinimumSectionSize(20)
        self.details_table.verticalHeader().setDefaultSectionSize(20)

        self.layout.addWidget(self.details_table, 3, 0, 1, 11)

        self.horizontal_spacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.layout.addItem(self.horizontal_spacer, 2, 8, 1, 1)

        self.a_currency = OptionalCurrencyComboBox(IncomeSpendingOperation)
        self.a_currency.setObjectName(u"a_currency")

        self.layout.addWidget(self.a_currency, 1, 7, 1, 1)

        self.currency = AccountCurrencyLabel(IncomeSpendingOperation)
        self.currency.setObjectName(u"currency")

        self.layout.addWidget(self.currency, 1, 6, 1, 1)


        self.retranslateUi(IncomeSpendingOperation)

        QMetaObject.connectSlotsByName(IncomeSpendingOperation)
    # setupUi

    def retranslateUi(self, IncomeSpendingOperation):
        IncomeSpendingOperation.setWindowTitle(QCoreApplication.translate("IncomeSpendingOperation", u"Form", None))
        self.account_label.setText(QCoreApplication.translate("IncomeSpendingOperation", u"Account", None))
        self.date_label.setText(QCoreApplication.translate("IncomeSpendingOperation", u"Date/Time", None))
#if QT_CONFIG(tooltip)
        self.add_button.setToolTip(QCoreApplication.translate("IncomeSpendingOperation", u"Add detail", None))
#endif // QT_CONFIG(tooltip)
        self.add_button.setText("")
#if QT_CONFIG(tooltip)
        self.del_button.setToolTip(QCoreApplication.translate("IncomeSpendingOperation", u"Remove detail", None))
#endif // QT_CONFIG(tooltip)
        self.del_button.setText("")
#if QT_CONFIG(tooltip)
        self.copy_button.setToolTip(QCoreApplication.translate("IncomeSpendingOperation", u"Copy detail", None))
#endif // QT_CONFIG(tooltip)
        self.copy_button.setText("")
        self.note_label.setText(QCoreApplication.translate("IncomeSpendingOperation", u"Note", None))
#if QT_CONFIG(tooltip)
        self.commit_button.setToolTip(QCoreApplication.translate("IncomeSpendingOperation", u"Commit changes", None))
#endif // QT_CONFIG(tooltip)
        self.commit_button.setText("")
        self.peer_label.setText(QCoreApplication.translate("IncomeSpendingOperation", u"Peer", None))
        self.timestamp_editor.setDisplayFormat(QCoreApplication.translate("IncomeSpendingOperation", u"dd/MM/yyyy hh:mm:ss", None))
        self.main_label.setText(QCoreApplication.translate("IncomeSpendingOperation", u"Income / Spending", None))
        self.details_label.setText(QCoreApplication.translate("IncomeSpendingOperation", u"Details", None))
#if QT_CONFIG(tooltip)
        self.revert_button.setToolTip(QCoreApplication.translate("IncomeSpendingOperation", u"Cancel changes", None))
#endif // QT_CONFIG(tooltip)
        self.revert_button.setText("")
        self.currency.setText(QCoreApplication.translate("IncomeSpendingOperation", u"CUR", None))
    # retranslateUi




from datetime import datetime
from dateutil import tz
from decimal import Decimal

from PySide6.QtCore import Qt, Slot, QByteArray
from PySide6.QtWidgets import QMessageBox
from jal.ui.widgets.ui_transfer_operation import Ui_TransferOperation
from jal.widgets.abstract_operation_details import AbstractOperationDetails
from jal.widgets.delegates import WidgetMapperDelegateBase
from jal.db.operations import LedgerTransaction
from jal.db.helpers import db_row2dict
from jal.db.account import JalAccount


# ----------------------------------------------------------------------------------------------------------------------
class TransferWidgetDelegate(WidgetMapperDelegateBase):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.delegates = {'withdrawal_timestamp': self.timestamp_delegate,
                          'withdrawal': self.decimal_delegate,
                          'deposit_timestamp': self.timestamp_delegate,
                          'deposit': self.decimal_delegate,
                          'fee': self.decimal_delegate}


# ----------------------------------------------------------------------------------------------------------------------
class TransferWidget(AbstractOperationDetails):
    def __init__(self, parent=None):
        super().__init__(parent=parent, ui_class=Ui_TransferOperation)
        self.name = self.tr("Transfer")
        self.operation_type = LedgerTransaction.Transfer

        self.ui.copy_date_btn.setFixedWidth(self.ui.copy_date_btn.fontMetrics().horizontalAdvance("XXXX"))
        self.ui.copy_amount_btn.setFixedWidth(self.ui.copy_amount_btn.fontMetrics().horizontalAdvance("XXXX"))
        self.ui.withdrawal_timestamp.setFixedWidth(self.ui.withdrawal_timestamp.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.25)
        self.ui.deposit_timestamp.setFixedWidth(self.ui.deposit_timestamp.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.25)
        self.ui.fee_account_widget.setValidation(False)
        self.ui.asset_widget.setValidation(False)

        self.ui.copy_date_btn.clicked.connect(self.onCopyDate)
        self.ui.copy_amount_btn.clicked.connect(self.onCopyAmount)
        self.ui.fee_check.clicked.connect(self.on_fee_check_change)
        self.ui.asset_check.clicked.connect(self.on_asset_check_change)

        super()._init_db("transfers")
        self.mapper.setItemDelegate(TransferWidgetDelegate(self.mapper))

        self.ui.from_account_widget.changed.connect(self.mapper.submit)
        self.ui.to_account_widget.changed.connect(self.mapper.submit)
        self.ui.fee_account_widget.changed.connect(self.mapper.submit)
        self.ui.asset_widget.changed.connect(self.mapper.submit)
        self.mapper.currentIndexChanged.connect(self.on_record_change)

        self.mapper.addMapping(self.ui.withdrawal_timestamp, self.model.fieldIndex("withdrawal_timestamp"))
        self.mapper.addMapping(self.ui.from_account_widget, self.model.fieldIndex("withdrawal_account"))
        self.mapper.addMapping(self.ui.from_currency, self.model.fieldIndex("withdrawal_account"))
        self.mapper.addMapping(self.ui.withdrawal, self.model.fieldIndex("withdrawal"))
        self.mapper.addMapping(self.ui.deposit_timestamp, self.model.fieldIndex("deposit_timestamp"))
        self.mapper.addMapping(self.ui.to_account_widget, self.model.fieldIndex("deposit_account"))
        self.mapper.addMapping(self.ui.to_currency, self.model.fieldIndex("deposit_account"))
        self.mapper.addMapping(self.ui.deposit, self.model.fieldIndex("deposit"))
        self.mapper.addMapping(self.ui.fee_account_widget, self.model.fieldIndex("fee_account"), QByteArray("selected_id_str"))
        self.mapper.addMapping(self.ui.fee_currency, self.model.fieldIndex("fee_account"))
        self.mapper.addMapping(self.ui.fee, self.model.fieldIndex("fee"))
        self.mapper.addMapping(self.ui.asset_widget, self.model.fieldIndex("asset"), QByteArray("selected_id_str"))
        self.mapper.addMapping(self.ui.number, self.model.fieldIndex("number"))
        self.mapper.addMapping(self.ui.note, self.model.fieldIndex("note"))

        self.model.select()

    def _validated(self):
        fields = db_row2dict(self.model, 0)
        # Set related fields NULL if we don't have fee. This is required for correct transfer processing
        if not fields['fee'] or Decimal(fields['fee']) == Decimal('0'):
            self.model.setData(self.model.index(0, self.model.fieldIndex("fee_account")), None)
            self.model.setData(self.model.index(0, self.model.fieldIndex("fee")), None)
        else:
            if fields['fee_account'] == '0':
                QMessageBox().warning(self, self.tr("Incomplete data"), self.tr("An account isn't chosen for fee collection from"), QMessageBox.Ok)
                return False
            if not JalAccount(int(fields['fee_account'])).organization():
                QMessageBox().warning(self, self.tr("Incomplete data"), self.tr("Can't collect fee from an account without organization assigned"), QMessageBox.Ok)
                return False
        if fields['asset'] == '0':   # Store None if asset isn't selected
            self.model.setData(self.model.index(0, self.model.fieldIndex("asset")), None)
        return True

    def revertChanges(self):
        super().revertChanges()
        self.on_record_change(0)

    def prepareNew(self, account_id):
        new_record = super().prepareNew(account_id)
        new_record.setValue("withdrawal_timestamp", int(datetime.now().replace(tzinfo=tz.tzutc()).timestamp()))
        new_record.setValue("withdrawal_account", account_id)
        new_record.setValue("withdrawal", '0')
        new_record.setValue("deposit_timestamp", int(datetime.now().replace(tzinfo=tz.tzutc()).timestamp()))
        new_record.setValue("deposit_account", 0)
        new_record.setValue("deposit", '0')
        new_record.setNull("fee_account")
        new_record.setValue("fee", '0')
        new_record.setNull("asset")
        new_record.setValue("number", None)
        new_record.setValue("note", None)
        return new_record

    def copyToNew(self, row):
        new_record = self.model.record(row)
        new_record.setNull("id")
        new_record.setValue("withdrawal_timestamp", int(datetime.now().replace(tzinfo=tz.tzutc()).timestamp()))
        new_record.setValue("deposit_timestamp", int(datetime.now().replace(tzinfo=tz.tzutc()).timestamp()))
        return new_record

    @Slot()
    def onCopyDate(self):
        self.ui.deposit_timestamp.setDateTime(self.ui.withdrawal_timestamp.dateTime())
        # mapper.submit() isn't needed here as 'changed' signal of 'deposit_timestamp' is linked with it

    @Slot()
    def onCopyAmount(self):
        self.ui.deposit.setText(self.ui.withdrawal.text())
        self.mapper.submit()

    @Slot()
    def on_record_change(self, idx):
        if self.ui.fee_account_widget.selected_id:
            self.ui.fee_check.setCheckState(Qt.CheckState.Checked)
            self.set_fee_data_visible(True)
        else:
            self.ui.fee_check.setCheckState(Qt.CheckState.Unchecked)
            self.set_fee_data_visible(False)
        if self.ui.asset_widget.selected_id:
            self.ui.asset_check.setCheckState(Qt.CheckState.Checked)
            self.set_asset_data_visible(True)
        else:
            self.ui.asset_check.setCheckState(Qt.CheckState.Unchecked)
            self.set_asset_data_visible(False)

    def set_fee_data_visible(self, visible: bool):
        self.ui.fee_account_widget.setVisible(visible)
        self.ui.fee.setVisible(visible)
        self.ui.fee_currency.setVisible(visible)

    @Slot()
    def on_fee_check_change(self, _state):
        with_fee = self.ui.fee_check.isChecked()
        self.set_fee_data_visible(with_fee)
        if not with_fee:
            self.ui.fee_account_widget.selected_id = 0
            self.ui.fee.setText('')
        self.mapper.submit()

    def set_asset_data_visible(self, visible: bool):
        self.ui.asset_widget.setVisible(visible)
        self.ui.value_label.setVisible(visible)
        self.ui.copy_amount_btn.setVisible(not visible)
        self.ui.from_currency.setVisible(not visible)

    @Slot()
    def on_asset_check_change(self, _state):
        asset_transfer = self.ui.asset_check.isChecked()
        self.set_asset_data_visible(asset_transfer)
        if not asset_transfer:
            self.ui.asset_widget.selected_id = 0
        self.mapper.submit()



from functools import partial

from PySide6.QtCore import Qt, Slot, Signal, QDateTime, QSortFilterProxyModel
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QMenu, QMessageBox, QDialog
from jal.ui.ui_operations_widget import Ui_OperationsWidget
from jal.widgets.mdi import MdiWidget
from jal.widgets.selection_dialog import SelectTagDialog
from jal.db.helpers import load_icon
from jal.db.account import JalAccount
from jal.db.asset import JalAsset
from jal.db.balances_model import BalancesModel
from jal.db.operations_model import OperationsModel
from jal.db.operations import LedgerTransaction


# ----------------------------------------------------------------------------------------------------------------------
class OperationsWidget(MdiWidget):
    dbUpdated = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = Ui_OperationsWidget()
        self.ui.setupUi(self)

        self.current_index = None  # this is used in onOperationContextMenu() to track item for menu

        # Set icons
        self.ui.NewOperationBtn.setIcon(load_icon("new.png"))
        self.ui.CopyOperationBtn.setIcon(load_icon("copy.png"))
        self.ui.DeleteOperationBtn.setIcon(load_icon("delete.png"))

        # Operations view context menu
        self.contextMenu = QMenu(self.ui.OperationsTableView)
        self.actionReconcile = QAction(load_icon("reconcile.png"), self.tr("Reconcile"), self)
        self.actionTag = QAction(load_icon("tag.png"), self.tr("Assign tag"), self)
        self.actionCopy = QAction(load_icon("copy.png"), self.tr("Copy"), self)
        self.actionDelete = QAction(load_icon("delete.png"), self.tr("Delete"), self)
        self.contextMenu.addAction(self.actionReconcile)
        self.contextMenu.addSeparator()
        self.contextMenu.addAction(self.actionTag)
        self.contextMenu.addSeparator()
        self.contextMenu.addAction(self.actionCopy)
        self.contextMenu.addAction(self.actionDelete)

        # Customize UI configuration
        self.balances_model = BalancesModel(self.ui.BalancesTableView)
        self.ui.BalancesTableView.setModel(self.balances_model)
        self.balances_model.configureView()

        self.operations_model = OperationsModel(self.ui.OperationsTableView)
        self.operations_filtered_model = QSortFilterProxyModel(self.ui.OperationsTableView)
        self.operations_filtered_model.setSourceModel(self.operations_model)
        self.operations_filtered_model.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.ui.OperationsTableView.setModel(self.operations_filtered_model)
        self.operations_model.configureView()
        self.ui.OperationsTableView.setContextMenuPolicy(Qt.CustomContextMenu)

        self.connect_signals_and_slots()

        self.NewOperationMenu = QMenu()
        self.ui.OperationsTabs.dbUpdated.connect(self.dbUpdated)
        self.ui.OperationsTabs.dbUpdated.connect(self.operations_model.refresh)
        for key, name in self.ui.OperationsTabs.get_operations_list().items():
            self.NewOperationMenu.addAction(name, partial(self.create_operation, key))
        self.ui.NewOperationBtn.setMenu(self.NewOperationMenu)

        # Setup balance and holdings parameters
        current_time = QDateTime.currentDateTime()
        current_time.setTimeSpec(Qt.UTC)  # We use UTC everywhere so need to force TZ info
        self.ui.BalanceDate.setDateTime(current_time)
        self.ui.BalancesCurrencyCombo.setIndex(JalAsset.get_base_currency())

        self.ui.OperationsTableView.selectRow(0)
        self.ui.DateRange.setCurrentIndex(0)

    def connect_signals_and_slots(self):
        self.actionReconcile.triggered.connect(self.reconcile_at_current_operation)
        self.ui.BalanceDate.dateChanged.connect(self.ui.BalancesTableView.model().setDate)
        self.ui.BalancesCurrencyCombo.changed.connect(self.ui.BalancesTableView.model().setCurrency)
        self.ui.BalancesTableView.doubleClicked.connect(self.balance_double_click)
        self.ui.ShowInactiveCheckBox.stateChanged.connect(self.ui.BalancesTableView.model().toggleActive)
        self.ui.DateRange.changed.connect(self.operations_model.setDateRange)
        self.ui.ChooseAccountBtn.changed.connect(self.operations_model.setAccount)
        self.ui.SearchString.editingFinished.connect(self.update_operations_filter)
        self.ui.OperationsTableView.selectionModel().selectionChanged.connect(self.operation_selection_change)
        self.operations_model.modelReset.connect(partial(self.operation_selection_change, None, None))  # Use the same slot with NULL-parameters to clean operation view
        self.ui.OperationsTableView.customContextMenuRequested.connect(self.operation_context_menu)
        self.ui.DeleteOperationBtn.clicked.connect(self.delete_operation)
        self.actionDelete.triggered.connect(self.delete_operation)
        self.ui.CopyOperationBtn.clicked.connect(self.ui.OperationsTabs.copy_operation)
        self.actionCopy.triggered.connect(self.ui.OperationsTabs.copy_operation)
        self.actionTag.triggered.connect(self.assign_tag)

    @Slot()
    def delete_operation(self):
        if QMessageBox().warning(None, self.tr("Confirmation"),
                                 self.tr("Are you sure to delete selected transaction(s)?"),
                                 QMessageBox.Yes, QMessageBox.No) == QMessageBox.No:
            return
        rows = []
        for index in self.ui.OperationsTableView.selectionModel().selectedRows():
            rows.append(index.row())
        self.operations_model.delete_rows(rows)
        self.dbUpdated.emit()

    @Slot()
    def create_operation(self, operation_type):
        self.ui.OperationsTabs.new_operation(operation_type, self.operations_model.getAccount())

    @Slot()
    def update_operations_filter(self):
        self.ui.OperationsTableView.model().setFilterFixedString(self.ui.SearchString.text())
        self.ui.OperationsTableView.model().setFilterKeyColumn(-1)

    @Slot()
    def balance_double_click(self, index):
        self.ui.ChooseAccountBtn.account_id = index.model().getAccountId(index.row())

    @Slot()
    def operation_selection_change(self, selected, _deselected):
        op_type = LedgerTransaction.NA
        op_id = 0
        if len(self.ui.OperationsTableView.selectionModel().selectedRows()) == 1:
            idx = selected.indexes()
            if idx:
                selected_row = self.operations_filtered_model.mapToSource(idx[0]).row()
                op_type, op_id = self.operations_model.get_operation(selected_row)
        self.ui.OperationsTabs.show_operation(op_type, op_id)

    @Slot()
    def operation_context_menu(self, pos):
        self.current_index = self.ui.OperationsTableView.indexAt(pos)
        if len(self.ui.OperationsTableView.selectionModel().selectedRows()) != 1:
            self.actionReconcile.setEnabled(False)
            self.actionCopy.setEnabled(False)
        else:
            self.actionReconcile.setEnabled(True)
            self.actionCopy.setEnabled(True)
        self.contextMenu.popup(self.ui.OperationsTableView.viewport().mapToGlobal(pos))

    @Slot()
    def reconcile_at_current_operation(self):
        idx = self.operations_model.index(self.current_index.row(), 0)  # we need only row to address fields by name
        timestamp = self.operations_model.data(idx, Qt.UserRole, field="timestamp")
        account_id = self.operations_model.data(idx, Qt.UserRole, field="account_id")
        JalAccount(account_id).reconcile(timestamp)
        self.operations_model.refresh()

    def refresh(self):
        self.balances_model.update()

    @Slot()
    def assign_tag(self):
        rows = []
        for index in self.ui.OperationsTableView.selectionModel().selectedRows():
            rows.append(index.row())
        dialog = SelectTagDialog(parent=self, description=self.tr("Choose tag to be assigned to selected operations:"))
        if dialog.exec() != QDialog.Accepted:
            return
        self.operations_model.assign_tag_to_rows(rows, dialog.selected_id)
        self.dbUpdated.emit()
